/* libspf - Sender Permitted From library
*	
*  ANSI C implementation of draft-mengwong-spf-02.9.7.txt
*
*  Author: James Couzens <jcouzens@obscurity.org>
*  Author: Sean Comeau <scomeau@obscurity.org>
*
*  File:   dns.c
*  Desc:   DNS related functions
*
*  License:
*
*  The libspf Software License, Version 1.0
*
*  Copyright (c) 2004 James Couzens & Sean Comeau  All rights
*  reserved.
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*  1. Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in
*     the documentation and/or other materials provided with the
*     distribution.
*
*  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
*  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
*  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
*  DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
*  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
*  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
*  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
*  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
*  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
*  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
*  SUCH DAMAGE.
* 
*/

#include "spf.h"
#include "util.h"
#include "dns.h"
#include "error.h"

/* DNS_txt_answer
*
*  Author: James Couzens <jcouzens@obscurity.org>
*
*  Date:   12/10/03
*
*  Desc:   
*          Executes a DNS query of type T_TYPE and then calls the
*  appropriate answer parsing function based on that type.  Returns
*  a pointer to allocated memory (a string of space delimited 
*  records).  Upon failure returns NULL.
*  
*/
char *DNS_query(peer_info_t *peer_info, const char *s, const int T_TYPE,
  const char *mta)
{
  int16_t   r_len;                 /* res_search return code & packet len */
  int16_t   rc;                    /* generic return code / length of */
 
  u_char    *msg_ptr;              /* pointer to beginning of the message */
  u_char    *eom_ptr;              /* pointer to the end of the message */

  u_char    *rd_ptr;               /* pointer to uncompressed message */ 
  
  int8_t    ancount;               /* number of answers */

  HEADER    *hdr;                  /* pointer to the header of the packet */
  
  char      buf[SPF_MAXDNAME];     /* record extraction buffer */
  char      answer[SPF_PACKETSZ];  /* answer buffer */

  char      *rr_data = NULL;       /* record */
  
  if (s == NULL)
  {
    xeprintf("Passed a NULL char.  Aborting.\n");    
    return(NULL);
  }

  xvprintf("Called with (%s) and type: %i\n", s, T_TYPE);
  
  memset(answer, '\0', SPF_PACKETSZ);
  memset(buf, '\0', SPF_MAXDNAME);
  
  r_len = res_search(s, C_IN, T_TYPE, (u_char *)answer, SPF_PACKETSZ);
  
  if (r_len <= 0)
  {
    switch (h_errno)
    {
      case HOST_NOT_FOUND:
        snprintf(peer_info->error, MAX_ERROR, "DNS failure (Host not found)");
        UTIL_assoc_prefix(peer_info, SPF_NONE, NULL);
        xvprintf("%s", peer_info->error);
        /*xeprintf(peer_info->error);*/
        return(NULL);        

      case TRY_AGAIN:
        snprintf(peer_info->error, MAX_ERROR, "DNS failure (Non-Authoritative Host not "
          "found, or SERFAIL.)");
        UTIL_assoc_prefix(peer_info, SPF_NONE, NULL);
        xvprintf("%s", peer_info->error);
        /*xeprintf(peer_info->error);*/
        return(NULL);

      case NO_RECOVERY:
        snprintf(peer_info->error, MAX_ERROR, "DNS failure (Non recoverable errors, "
          "FORMERR, REFUSED, NOTIMP.)");
        UTIL_assoc_prefix(peer_info, SPF_ERROR, NULL);
        xvprintf("%s", peer_info->error);
        /*xeprintf(peer_info->error);*/
        return(NULL); 

      case NO_DATA:
        snprintf(peer_info->error, MAX_ERROR, "DNS failure (Valid name, no data record "
          "of requested, type.)");
        UTIL_assoc_prefix(peer_info, SPF_NONE, NULL);
        xvprintf("%s", peer_info->error);
        /*xeprintf(peer_info->error);*/
        return(NULL);

       default:
        snprintf(peer_info->error, MAX_ERROR, "DNS failure (internal error)");  
        UTIL_assoc_prefix(peer_info, SPF_ERROR, NULL);
        xvprintf("%s", peer_info->error);
        /*xeprintf(peer_info->error);*/
        return(NULL);
    } /* switch */
  }

  hdr = (HEADER *)&answer;
  ancount = ntohs(hdr->ancount);
 
  xvprintf("Received packet size of %i bytes which contains %i answers.\n",
    r_len, ancount);
  xvprintf("ANSWERS: %i\n", ancount);   
  xvprintf("QUESTIONS: %i\n", ntohs(hdr->qdcount));
  
  msg_ptr = (u_char *)&answer;            /* point to start of message */
  eom_ptr = (u_char *)&answer + r_len;    /* point to end of message */
  rd_ptr = (u_char *)&answer + HFIXEDSZ;  /* point to start of RDATA */
    
  /* Comment: James Couzens <jcouzens@obscurity.org>
  *  Date: 12/10/03
  *
  * This should be changed to a loop in the event that I am wrong 
  * with my understanding that because we are only making ONE question
  * to the server, for a single T_TXT record, we'll only ever need to 
  * parse one question out of the received packet(s). 
  *
  */

  if ((rc = dn_skipname(rd_ptr, eom_ptr)) < 0)
  {
    xeprintf("Error obtaining QUESTION!\n");
    return(NULL);
  }

  rd_ptr += rc + QFIXEDSZ;          /* jump to start of ANSWER */

  switch (T_TYPE)
  {
    case T_A:
      return((char *)TRUE);
      
    case T_TXT:
      if ((rr_data = DNS_txt_answer(ancount, (u_char *)msg_ptr, (u_char *)eom_ptr,
        (u_char *)rd_ptr, buf)) == NULL)
      {
        return(NULL);
      }
      break;
      
    case T_MX:
      if ((rr_data = DNS_mx_answer(ancount, (u_char *)msg_ptr, (u_char *)eom_ptr,
        (u_char *)rd_ptr, buf)) == NULL)
      {
        return(NULL);
      }
      break;
      
    case T_PTR:

  /* Comment: James Couzens <jcouzens@obscurity.org>
  *  Date: 01/04/04
  *
  *  DNS_ptr_answer doesn't allocate any memory and simply returns TRUE or FALSE,
  *  however, this function returns a char so we return a cast TRUE or FALSE
  *  depending on the outcome.
  *
  */
      if (DNS_ptr_answer(peer_info, ancount, (u_char *)msg_ptr, (u_char *)eom_ptr,
        (u_char *)rd_ptr, buf, mta) == FALSE)
      {
        return((char *)FALSE);
      }
      else
      {
        return((char *)TRUE);
      }
      break;
    case T_CNAME:
      if ((rr_data = DNS_cname_answer(ancount, (u_char *)msg_ptr, (u_char *)eom_ptr,
        (u_char *)rd_ptr, buf)) == NULL)
      {
        return(NULL);
      }
      break;
  } 
  
  return(rr_data);
}


/* DNS_txt_answer
*
*  Author: James Couzens <jcouzens@obscurity.org>
*
*  Date:   01/02/04
*  Date:   02/23/04 - Bugfix from Albert Weichselbraun <albert@atnet.at>
*
*  Desc:   
*          SPF_PACKETSZ bytes are allocated and then filled with \0 chars.
*  This buffer is then used in a TXT DNS query using data from the passed 
*  peer_info_t structure.  Upon success this buffer is re-cast as a char *
*  and then a pointer to this memory is returned.  Upon failure a NULL
*  pointer is returned.
*  
*/
char *DNS_txt_answer(int16_t ancount, const u_char *msg_ptr,
  const u_char *eom_ptr, u_char *rd_ptr, char *buf)
{
  int16_t   i;                /* utility */
  int16_t   rc;               /* generic return code / length of */
  int16_t   rd_type;          /* answer type */
  int16_t   rd_len;           /* res_search return code & packet len */

  char      *rr_data = NULL;  /* data pointer */
  
  if (msg_ptr == NULL || eom_ptr == NULL || rd_ptr == NULL || buf == NULL)
  {
    xeprintf("Called with NULL pointers\n");
    return(NULL);
  }
  
  i = 0;
  while (ancount > 0 && rd_ptr < eom_ptr)
  {
    if ((rc = dn_expand(msg_ptr, eom_ptr, rd_ptr, buf, SPF_MAXCDNAME)) < 0) 
    {
      xeprintf("Error obtaining ANSWER!\n");
      return(NULL);
    }
    
    /* dname, type, class, TTL, rdata len, rdata */
    rd_ptr += rc;               /* jump to start of ANSWER data */
    GETSHORT(rd_type, rd_ptr);  /* get response type */
    rd_ptr += INT16SZ;          /* skip class */
    rd_ptr += INT32SZ;          /* skip TTL */
    GETSHORT(rd_len, rd_ptr);   /* get data length */

    if (rd_type != T_TXT)
    {
      xvprintf("Ignoring record not of T_TXT type. (%i)\n", rd_type);
      rd_ptr += rd_len;
      continue;
    }

    rd_ptr++;                  /* skip  */
    rd_ptr[rd_len] = '\0';
    rd_ptr[rd_len - 1] = ' ';
    i += rd_len;
    
    xvprintf("Answer %i has length %i. (%i)\n", 
      ancount, rd_len, i);

    xvprintf("Answer Data: (%s) len: %i\n", rd_ptr, rd_len);
    
    if (rd_len <= SPF_MAXDNAME)
    {  
      if (rr_data == NULL)
      {
        rr_data = xmalloc(i + 1);
        memset(rr_data, '\0', (i + 1));
      }
      else
      {
        rr_data = xrealloc(rr_data, (i + 1));
      }

      xvprintf("REALLOCATE memory: %i bytes\n", i); 
      
      strncat(rr_data, (char *)rd_ptr, rd_len);        
      rr_data[i - 1] = ' ';
      rr_data[i] = '\0';
    }

    rd_ptr += rd_len;
    ancount--;
  }

  if (rr_data == NULL)
  {
    return(NULL);
  }
  
  rr_data[i] = '\0';
  return(rr_data);
}


/* DNS_mx_answer
*
*  Author: James Couzens <jcouzens@obscurity.org>
*
*  Date:   01/02/04
*
*  Desc:   
*          SPF_PACKETSZ bytes are allocated and then filled with \0 chars.
*  This buffer is then used in an MX DNS query using data from the passed 
*  peer_info_t structure.  Upon success this buffer is re-cast as a char *
*  and then a pointer to this memory is returned.  Upon failure a NULL
*  pointer is returned.
*  
*/
char *DNS_mx_answer(int16_t ancount, const u_char *msg_ptr,
  const u_char *eom_ptr, u_char *rd_ptr, char *buf)
{
  int16_t   i;                /* utility */
  int16_t   rc;               /* generic return code / length of */
  int16_t   rd_len = 0;       /* res_search return code & packet len */
  int16_t   rd_pref;          /* MX preference */  
  int16_t   rd_type;          /* answer type */
  size_t    buf_len;          /* buffer length */
  
  char      *rr_data = NULL;  /* data pointer */
  
  i = 0;
  while ((ancount > 0) && (rd_ptr < eom_ptr))
  {
    if ((rc = dn_expand(msg_ptr, eom_ptr, rd_ptr, buf, SPF_MAXCDNAME)) < 0) 
    {
      xeprintf("Error obtaining ANSWER!\n");
      return(NULL);
    }
    
    /* dname, type, class, TTL, rdata len, rdata */
    rd_ptr += rc;               /* jump to start of ANSWER data */
    GETSHORT(rd_type, rd_ptr);  /* get response type */
    rd_ptr += INT16SZ;          /* skip class */
    rd_ptr += INT32SZ;          /* skip TTL */
    GETSHORT(rd_len, rd_ptr);   /* get data length */
    
    if (rd_type != T_MX)
    {
      xprintf("Forged packet?!  We requested T_MX (15) but got %i\n", rd_type);
      rd_ptr += rd_len;
      continue;
    }

    GETSHORT(rd_pref, rd_ptr);  /* get MX preference */

    if ((rc = dn_expand(msg_ptr, eom_ptr, rd_ptr, buf, SPF_MAXCDNAME)) < 0) 
    {
      xeprintf("Error obtaining ANSWER!\n");
      return(NULL);
    } 

    xvprintf("MX: %s Preference: %i\n", buf, rd_pref);

    buf_len = strlen(buf);
    i += (buf_len + 1);
    
    if (rd_len <= SPF_MAXDNAME)
    {  
      if (rr_data == NULL)
      {
        rr_data = xmalloc(i + 1);
        memset(rr_data, '\0', (i + 1));
      }
      else
      {
        rr_data = xrealloc(rr_data, (i + 1));
      }

      xvprintf("REALLOCATE memory: %i bytes\n", (i + 1));

      strncat(rr_data, buf, buf_len);
      rr_data[i - 1] = ' ';
      rr_data[i] = '\0';
    }
    
    rd_ptr += rc;
    ancount--;
  }
  
  rr_data[i - 1] = '\0';
  return(rr_data);
}


/* DNS_ptr_answer
*
*  Author: James Couzens <jcouzens@obscurity.org>
*
*  Date:   01/02/04
*
*  Desc:   
*          A reverse lookup on an IP address leads to a lookup per returned
*  PTR answer to see if the returned answer matches.  The forward lookups are
*  handled by a separate function which calls gethostbyname.  Upon a single 
*  successful match of a forward lookup with a reverse lookup, returns TRUE.
*  Returns FALSE upon failure.
*  
*/
SPF_BOOL DNS_ptr_answer(peer_info_t *peer_info, int16_t ancount, 
const u_char *msg_ptr, const u_char *eom_ptr, u_char *rd_ptr, char *buf,
  const char *mta)
{
  int16_t   rc;                 /* generic return code / length of */
  int16_t   rd_len = 0;         /* res_search return code & packet len */
  int16_t   rd_type;            /* answer type */

  char      *cp       = NULL;   /* working pointer for copy of mta */
  char      *copy     = NULL;   /* copy of mta */
  char      *ptr_dom  = NULL;   /* stores domain from buf's (ptr res) hostname */
  char      *cur_dom  = NULL;   /* stores domain from rpeer's hostname */
  
  while ((ancount > 0) && (rd_ptr < eom_ptr))
  {
    if ((rc = dn_expand(msg_ptr, eom_ptr, rd_ptr, buf, SPF_MAXCDNAME)) < 0) 
    {
      xeprintf("Error obtaining ANSWER!\n");
      return(FALSE);
    }

     /* dname, type, class, TTL, rdata len, rdata */
    rd_ptr += rc;               /* jump to start of ANSWER data */
    GETSHORT(rd_type, rd_ptr);  /* get response type */
    rd_ptr += INT16SZ;          /* skip class */
    rd_ptr += INT32SZ;          /* skip TTL */
    GETSHORT(rd_len, rd_ptr);   /* get data length */

    if (rd_type != T_PTR)
    {
      xvprintf("Forged packet?!  We requested T_PTR (12) but got %i\n",
      rd_type);
      rd_ptr += rd_len;
      continue;
    }
    
    if ((rc = dn_expand(msg_ptr, eom_ptr, rd_ptr, buf, SPF_MAXCDNAME)) < 0) 
    {
      xeprintf("Error obtaining ANSWER!\n");
      return(FALSE);
    }

    xvprintf("Answer %i has length %i.\n", ancount, rd_len);
    xvprintf("Answer Data: (%s) len: %i\n", buf, strlen(buf));
    
    if (rd_len <= SPF_MAXDNAME)
    {
      if ((ptr_dom = UTIL_get_dname(buf)) == NULL)
      {
        xvprintf("Invalid hostname (%s)\n", buf);
        return(FALSE);
      }

      cp = copy = xstrndup(mta, (strlen(mta) + 1));
      
      if ((cp = strstr(mta, ":")) != NULL)
      {
        cp++;
      }
      else
      {
        cp = (char *)mta; /* stfu compiler */
      }
      
      if ((cur_dom = UTIL_get_dname(cp)) == NULL)
      {
        xvprintf("Invalid hostname (%s)\n", mta);
        xfree(ptr_dom); 
        xfree(copy);
        return(FALSE);
      }

      if (UTIL_validate_hostname(peer_info, buf, 32) == FALSE)
      {
        xvprintf("Unable to validate hostname (%s [%s]) with (%s [%s])\n", 
          buf, ptr_dom, mta, cur_dom);
          
        xfree(ptr_dom);
        xfree(cur_dom);
        xfree(copy);
        return(FALSE);
      }

      if (strcasecmp(ptr_dom, cur_dom) == 0)
      {
        xprintf("Validated (%s [%s]) against(%s [%s])\n", buf, ptr_dom,
          mta, cur_dom);

        peer_info->r_vhname = xstrndup(mta, (strlen(mta) + 1));
        xfree(ptr_dom);
        xfree(cur_dom);
        xfree(copy);
        return(TRUE);
      }
      else
      {
        xprintf("Unable to validate (%s : %s)\n", buf, mta);
        xfree(ptr_dom);
        xfree(cur_dom);
        xfree(copy);
      }
    }
    else
    {
      xeprintf("Answer length is too long!\n");
    }

    rd_ptr += rc;
    ancount--;
  }
 
  return(FALSE); 
}

/* DNS_cname_answer
*
*  Author: Teddy <teddy@teddy.ch>
*
*  Date:   29/04/04
*
*  Desc:   
*          SPF_PACKETSZ bytes are allocated and then filled with \0 chars.
*  This buffer is then used in a TXT DNS query using data from the passed 
*  peer_info_t structure.  Upon success this buffer is re-cast as a char *
*  and then a pointer to this memory is returned.  Upon failure a NULL
*  pointer is returned.
*  
*/
char *DNS_cname_answer(int16_t ancount, const u_char *msg_ptr,
  const u_char *eom_ptr, u_char *rd_ptr, char *buf)
{
  int16_t   i;                /* utility */
  int16_t   rc;               /* generic return code / length of */
  int16_t   rd_type;          /* answer type */
  int16_t   rd_len;           /* res_search return code & packet len */
  size_t    buf_len;          /* buffer length */

  char      *rr_data = NULL;  /* data pointer */
  
  if (msg_ptr == NULL || eom_ptr == NULL || rd_ptr == NULL || buf == NULL)
  {
    xeprintf("Called with NULL pointers\n");
    return(NULL);
  }

  i = 0;
  while (ancount > 0 && rd_ptr < eom_ptr)
  {
    if ((rc = dn_expand(msg_ptr, eom_ptr, rd_ptr, buf, SPF_MAXCDNAME)) < 0) 
    {
      xeprintf("Error obtaining ANSWER!\n");
      return(NULL);
    }
    
    /* dname, type, class, TTL, rdata len, rdata */
    rd_ptr += rc;               /* jump to start of ANSWER data */
    GETSHORT(rd_type, rd_ptr);  /* get response type */
    rd_ptr += INT16SZ;          /* skip class */
    rd_ptr += INT32SZ;          /* skip TTL */
    GETSHORT(rd_len, rd_ptr);   /* get data length */

    if (rd_type != T_CNAME)
    {
      xvprintf("Ignoring record not of T_CNAME type. (%i)\n", rd_type);
      rd_ptr += rd_len;
      continue;
    }

    if ((rc = dn_expand(msg_ptr, eom_ptr, rd_ptr, buf, SPF_MAXCDNAME)) < 0) 
    {
      xeprintf("Error obtaining ANSWER!\n");
      return(NULL);
    } 

    buf_len = strlen(buf);
    i += (buf_len + 1);
    
    if (rd_len <= SPF_MAXDNAME)
    {  
      if (rr_data == NULL)
      {
        rr_data = xmalloc(i + 1);
        memset(rr_data, '\0', (i + 1));
      }
      else
      {
        rr_data = xrealloc(rr_data, (i + 1));
      }

      xvprintf("REALLOCATE memory: %i bytes\n", (i + 1));

      strncat(rr_data, buf, buf_len);
      rr_data[i - 1] = ' ';
      rr_data[i] = '\0';
    }
    
    rd_ptr += rc;
    ancount--;
  }
  
  rr_data[i - 1] = '\0';
  return(rr_data);
}


/* end dns.c */
